from __future__ import annotations
from jellybeans.structures import Graph, Queue, PriorityQueue, UFDS
from .private_funcs import (
    _BFS,
    _path_construction,
    _dfs_topo,
    _dfs, 
    _initializer,
    _dfs_sssp_tree,
    _bfs_sssp_unweighted,
    _relax,
    _floyd_sp,
    _floyd_reachability,
    _floyd_detect_cycle
)
from jellybeans.exceptions.negative_cycle import _Negativecycle


def reachability(graph: Graph, source: int, destination: int) -> tuple:
    '''
    This function checks if it is possible to travel from a source vertex to a destination vertex 

    Args:
        graph: Graph object
        source: Source vertex
        destination: Destination vertex
    Returns:
        A tuple consisting of a boolean value and the path to travel from source to destination.
    '''
    visited, parent, mapping = _initializer(True, True, True, graph)
    adj_list = graph.to_adj_list()
    _BFS(visited, parent, mapping, source, adj_list)
    return (True, _path_construction(parent, mapping, source, destination)) if visited[mapping[destination]] == 1 else (False, None)


def counting_components(graph: Graph) -> int:
    '''
    This function checks for the number of components in an UNDIRECTED graph. 

    Args:
        graph: Graph object
    Returns:
        Number of connected components
    '''
    components = 0
    visited, parent, mapping = _initializer(True, True, True, graph)
    vertices = graph.list_vertices()
    adj_list = graph.to_adj_list()
    for v in vertices:
        if visited[mapping[v]] == 0:
            components += 1
            _BFS(visited, parent, mapping, v, adj_list)
    return components


def topological_sort(graph: Graph) -> list:
    '''
    Performs a topological sort on the directed graph. Based on Kahn's Algorithm 

    Args:
        graph: Graph object
    Returns:
        List containing a valid topological ordering
    '''
    in_degree = []
    mapping = {}
    toposort = []
    vertices = graph.list_vertices()
    inv_map = [0 for _ in vertices]
    adj_list = graph.to_adj_list()
    counter = 0
    q = Queue()
    for v in vertices:
        in_degree.append(0)
        mapping[v] = counter
        inv_map[counter] = v
        counter += 1
    for _, vTo, _ in graph.to_edge_list():
        in_degree[mapping[vTo]] += 1

    for idx in range(len(in_degree)):
        if in_degree[idx] == 0:
            q.enqueue(inv_map[idx])

    while not q.is_empty():
        tex = q.dequeue()
        toposort.append(tex)
        for v, _ in adj_list[tex]:
            in_degree[mapping[v]] -= 1
            if in_degree[mapping[v]] == 0:
                q.enqueue(v)
    return toposort


def dfs_toposort(graph: Graph) -> list:
    '''
    Performs a topological sort on the directed graph. This is a DFS implementation 

    Args:
        graph: Graph object
    Returns:
        List containing a valid topological ordering
    '''
    vertices = graph.list_vertices()
    toposort = []
    visited, _, mapping = _initializer(True, False, True, graph)
    for v in vertices:
        if visited[mapping[v]] == 0:
            _dfs_topo(visited, toposort, v, graph.to_adj_list(), mapping)
    toposort.reverse()
    return toposort


def count_strong_connected_components(graph: Graph) -> int:
    '''
    Finds the number of strongly connected components (SCC) in a directed graph 

    Args:
        graph: Graph object
    Returns:
        Number of SCCs
    '''
    toposort = dfs_toposort(graph)
    SCC = 0
    visited, _, mapping = _initializer(True, False, True, graph)
    for idx in range(len(toposort)):
        ele = toposort[idx]
        if visited[mapping[ele]] == 0:
            SCC += 1
            _dfs(visited, ele, graph.transpose().to_adj_list(), mapping)
    return SCC


def spanning_tree_prim(graph: Graph, source: int, minimum: bool) -> Graph:
    '''
    Finds the minimum/maximum spanning tree of a graph using Prim's Algorithm 

    Args:
        graph: Graph Object
        source: Source vertex to begin algorithm
        minimum: True=minimum, False=Maximum
    Returns:
        Graph object of the mst
    '''
    adj_list = graph.to_adj_list()
    mst = Graph()
    visited, _, mapping = _initializer(True, True, True, graph)
    pq = PriorityQueue(comparator=(
        lambda x, y: x[2] <= y[2]) if minimum else (lambda x, y: x[2] >= y[2]))
    for v in graph.list_vertices():
        mst.add_vertex(v)

    for vTo, weight in adj_list[source]:
        pq.insert((source, vTo, weight))
    visited[source] = 1
    while not pq.is_empty():
        vFrom, vTo, weight = pq.extract()
        if visited[mapping[vTo]] == 0:
            mst.add_bidirected_edge(vFrom, vTo, (weight, weight))
            visited[mapping[vTo]] = 1
            for to, w in adj_list[vTo]:
                if visited[mapping[to]] == 0:
                    pq.insert((vTo, to, w))
    return mst


def spanning_tree_kruskal(graph: Graph, minimum: bool) -> Graph:
    '''
    Finds the minimum/maximum spanning tree of a graph using Kruskal's Algorrithm 

    Args:
        graph: Graph Object
        minimum: True=minimum, False=Maximum
    Returns:
        Graph object of the mst
    '''
    edge_list_sorted = sorted(
        graph.to_edge_list(), key=lambda x: x[2], reverse=not minimum)
    counter = 1
    mapping = {}
    mst = Graph()
    ufds = UFDS(graph.num_vertices())

    for v in graph.list_vertices():
        mst.add_vertex(v)
        mapping[v] = counter
        counter += 1
    for vFrom, vTo, weight in edge_list_sorted:
        if not ufds.is_same_set(mapping[vFrom], mapping[vTo]):
            ufds.union(mapping[vFrom], mapping[vTo])
            mst.add_bidirected_edge(vFrom, vTo, (weight, weight))
    return mst


def sssp_tree(graph: Graph, source: int) -> dict:
    '''
    Finds the single source shortest path of a tree.
    
    Args:
        graph: Graph Object
        source: Source vertex Number
    Returns:
        A dictionary of vertex -> cost
    '''
    if not graph.is_tree():
        raise TypeError("This graph is not a tree")

    vertices = graph.list_vertices()
    cost = [1000000000 for _ in vertices]
    parent = [-1 for _ in vertices]
    mapping = {vertices: idx for idx, vertices in enumerate(vertices)}
    cost[mapping[source]] = 0
    _dfs_sssp_tree(source, parent, cost, graph.to_adj_list(), mapping)
    return {vertices[idx]: cost for idx, cost in enumerate(cost)}


def sssp_unweighted(graph: Graph, source: int) -> dict:
    '''
    Finds the single source shortest path of an unweighted graph.

    Args:
        graph: Graph Object
        source: Source vertex Number
    Returns:
        A dictionary of vertex -> cost
    '''
    if not graph.is_unweighted():
        raise TypeError("This graph is not unweighted")
    vertices = graph.list_vertices()
    cost = [1000000000 for _ in vertices]
    parent = [-1 for _ in vertices]
    mapping = {vertices: idx for idx, vertices in enumerate(vertices)}
    cost[mapping[source]] = 0
    _bfs_sssp_unweighted(source, parent, cost, graph.to_adj_list(), mapping)
    return {vertices[idx]: cost for idx, cost in enumerate(cost)}


def sssp_dag(graph: Graph, source: int) -> dict:
    '''
    Finds the single source shortest path of a Directed Acyclic Graph (DAG).

    Args:
        graph: Graph Object
        source: Source vertex Number
    Returns:
        A dictionary of vertex -> cost
    '''
    if not graph.is_dag():
        raise TypeError("This graph is not a DAG")
    vertices = graph.list_vertices()
    cost = [1000000000 for _ in vertices]
    parent = [-1 for _ in vertices]
    mapping = {vertices: idx for idx, vertices in enumerate(vertices)}
    cost[mapping[source]] = 0

    topo = topological_sort(graph)

    for v_from in topo:
        for neighbor, weight in graph.to_adj_list()[v_from]:
            _relax(mapping[v_from], mapping[neighbor], weight, parent, cost)

    return {vertices[idx]: cost for idx, cost in enumerate(cost)}


def sssp_bellman_ford(graph: Graph, source: int) -> dict:
    '''
    Find the single source shortest path of any weighted graph using the bellman ford algorithm.

    Args:
        graph: Graph Object
        source: Source vertex Number
    Returns:
        A dictionary of vertex -> cost
    '''
    vertices = graph.list_vertices()
    cost = [1000000000 for _ in vertices]
    parent = [-1 for _ in vertices]
    mapping = {vertices: idx for idx, vertices in enumerate(vertices)}
    cost[mapping[source]] = 0
    edge_list = graph.to_edge_list()

    for _ in range(len(vertices) - 1):
        for v_from, v_to, weight in edge_list:
            _relax(mapping[v_from], mapping[v_to], weight, parent, cost)

    for v_from, v_to, weight in edge_list:
        if cost[mapping[v_from]] != 1000000000 and cost[mapping[v_to]] > cost[mapping[v_from]] + weight:
            raise _Negativecycle("Negative Weight Cycle detected!")
    return {vertices[idx]: cost for idx, cost in enumerate(cost)}


def sssp_dijkstra(graph: Graph, source: int) -> dict:
    '''
    Find the single source shortest path of any graph with no negative weight edge.

    Args:
        graph: Graph Object
        source: Source vertex Number
    Returns:
        A dictionary of vertex -> cost
    '''
    if not graph.is_positive():
        raise TypeError("This graph contains negative weight edges")
    vertices = graph.list_vertices()
    pq = PriorityQueue(comparator=lambda x, y: x[0] <= y[0])
    costs = {}
    for v in vertices:
        if v == source:
            costs[v] = 0
            pq.insert([0, v])
        else:
            costs[v] = 1000000000
            pq.insert((1000000000, v))
    while not pq.is_empty():
        curr_weight, curr_vertex = pq.extract()

        for neighbor, weight in graph.to_adj_list()[curr_vertex]:
            if costs[neighbor] > curr_weight + weight:
                pq.update((costs[neighbor], neighbor),
                          (curr_weight + weight, neighbor))
                costs[neighbor] = curr_weight + weight
    return costs


def floyd_warshall(graph: Graph, type: int) -> tuple:
    '''
    Find the all pairs shortest path by using Floud Warshall algorithm

    Args:
        graph: Graph Object
        source: Source vertex Number
    Returns:
        A tuple containing 2 elements. (mapping, dictionary of vertex -> cost)
    '''
    if type == 1:
        return _floyd_sp(graph.list_vertices(), graph.to_adj_list())
    elif type == 2:
        return _floyd_reachability(graph.list_vertices(), graph.to_adj_list())
    elif type == 3:
        return _floyd_detect_cycle(graph.list_vertices(), graph.to_adj_list())
    else:
        raise TypeError(
            "Invalid parameter type: type should be one of 1, 2, 3")



from math import floor
from typing import Any, Callable
from jellybeans.exceptions.not_one_based_index import _NotOneBasedIndexed

class PriorityQueue:
    '''
    PriorityQueue is an implementation of the binary heap data structure
    Default: MaxHeap
    '''
    def __init__(self, arr:list = None, comparator:Callable = lambda x, y: x >=y) -> None:
        '''
        Args:
            list: list of elements to be included into the heap (One based index)
            comparator: Function used to decide the linear ordering of the elements
        '''
        if type(arr) != list and arr != None:
            raise Exception("parameter arr must be of type list")
        elif arr != None and (len(arr) == 0 or arr[0] != None):
            raise _NotOneBasedIndexed("Input is not a one based Indexed list")
        
        self.__size = 0 if arr is None else len(arr) - 1
        self.__container = [None] if arr is None else arr
        self.__comparator = comparator
        self.__pos = {} # Dictionary of elements to a set of indexes
        for x in range(1, len(self.__container)):
            self.__add_index(x)

        self.__quick_tidy()

    def __add_index(self, idx:int) -> None:
        '''
        INTERNAL FUNCTION

        This is a utility function for adding an item and its index into self.__pos.

        Args:
            idx: The index of the item to be added
        '''
        if str(self.__container[idx]) in self.__pos:
            sett = self.__pos[str(self.__container[idx])]
            sett.add(idx)
        else:
            self.__pos[str(self.__container[idx])] = {idx}

    def __update_index(self, old_idx:int, new_idx:int) -> None:
        '''
        INTERNAL FUNCTION

        This is a utility function for updating the index of an item in self.__pos.

        Args:
            old_idx: The original index of the item to be updated
            new_idx: The new index of the item to be updated
        '''
        if self.__container[old_idx] != self.__container[new_idx]:
            self.__pos[str(self.__container[new_idx])].remove(old_idx)
            self.__pos[str(self.__container[new_idx])].add(new_idx)

    def __remove_index(self, item:Any, index:int) -> None:
        '''
        INTERNAL FUNCTION

        This is a utility function for removing the index of an item in self.__pos.

        Args:
            item: The item to remove in self.__pos
            new_idx: The index to be removed
        '''
        if len(self.__pos[str(item)]) > 1:
            self.__pos[str(item)].remove(index)
        else:
            del self.__pos[str(item)]

    def __shift_up(self, idx:int) -> None:
        '''
        INTERNAL FUNCTION

        Conduct shiftups on a node so that heap property is not violated

        Args:
            idx: The index to conduct the shiftUp operation on
        '''
        while idx > 1 and self.__comparator(self.__container[idx], self.__container[self.__parent(idx)]):
            self.__container[idx], self.__container[self.__parent(idx)] = self.__container[self.__parent(idx)], self.__container[idx], 
            self.__update_index(self.__parent(idx), idx)
            self.__update_index(idx, self.__parent(idx))
            idx = self.__parent(idx)

    def __shift_down(self, idx:int) -> None:
        '''
        INTERNAL FUNCTION

        Conduct shiftdowns on a node so that heap property is not violated

        Args:
            idx: The index to conduct the shiftDown operation on
        '''
        while idx <= self.__size:
            currIdx = idx
            currNode = self.__container[idx]
            if self.__left(idx) <= self.__size and self.__comparator(self.__container[self.__left(idx)], currNode):
                currIdx = self.__left(idx)
                currNode = self.__container[self.__left(idx)]

            if self.__right(idx) <= self.__size and self.__comparator(self.__container[self.__right(idx)], currNode):
                currIdx = self.__right(idx)
                currNode = self.__container[self.__right(idx)]

            if (currIdx != idx):
                self.__container[idx], self.__container[currIdx] = self.__container[currIdx],  self.__container[idx]
                self.__update_index(currIdx, idx)
                self.__update_index(idx, currIdx)
                idx = currIdx
            else:
                break

    def __parent(self, i:int) -> int:
        '''
        INTERNAL FUNCTION

        Args:
            index of the current node
        Returns:
            parent index of the current node
        '''
        return floor(i/2)

    def __left(self, i:int) -> int:
        '''
        INTERNAL FUNCTION

        Args:
            index of the current node
        Returns:
            left child of the current node
        '''
        return i*2

    def __right(self, i:int) -> int:
        '''
        INTERNAL FUNCTION

        Args:
            index of the current node
        Returns:
            right child of the current node
        '''
        return i*2 + 1

    def __quick_tidy(self) -> None:
        for idx in range(self.__parent(self.__size), 0, -1):
            self.__shift_down(idx)
    
    def insert(self, item:Any) -> None:
        '''
        Insert an item into the heap

        Args:
            item: item to be inserted into the heap
        '''
        self.__size += 1
        self.__container.append(item)
        self.__add_index(self.__size)
        self.__shift_up(self.__size)

    def peek(self) -> Any:
        '''
        Take a look at the root node of the heap

        Returns:
            root node of the heap
        '''
        return self.__container[1]

    def extract(self) -> Any:
        '''
        Returns and remove the root node of the heap

        Rerturns:
            value of the root node of the heap
        '''
        result = self.__container[1]
        self.__container[1] = self.__container[self.__size]
        
        self.__add_index(1)
        
        self.__size -= 1
        self.__container.pop()
        self.__remove_index(result, 1)
        self.__shift_down(1)
        return result

    def sort(self) -> list:
        '''
        Returns a sorted list according to the rules of the comparator

        Returns:
            list of the sorted elements
        '''
        result = []
        for _ in range(self.__size):
            result.append(self.extract())
        return result

    def is_empty(self) -> bool:
        '''
        Checks if the binary heap is empty
        
        Returns:
            True if heap is empty
        '''
        return self.__size == 0

    def search(self, item:Any) -> bool:
        '''
        Check if an item is present in the heap

        Args:
            item: Item to be found
        Returns:
            True if item is present
        '''
        return str(item) in self.__pos

    def update(self, item:Any, new_item:Any) -> None:
        '''
        Replace an item in the binary heap

        Args:
            item: The item to be replaced
            new_item: The new item
        '''
        if str(item) not in self.__pos:
            raise KeyError(f"{item} not found in Heap")
        idx_item = min(self.__pos[str(item)])
        self.__container[idx_item] = new_item
        self.__remove_index(item, idx_item)
        self.__add_index(idx_item)
        self.__shift_up(idx_item)
        self.__shift_down(idx_item)
        return idx_item

    def __len__(self) -> int:
        '''
        Returns the size of the heap
        '''
        return self.__size

    def __str__(self) -> str:
        '''
        Returns a string representation of the heap
        '''
        return str(self.__container.copy())


class UFDS:
    '''
    Union Find Disjoint Sets
    '''
    def __init__(self, sets:int) -> None:
        '''
        Args:
            Number of disjoint sets to create
        '''
        self.__parent = [None]
        self.__movements = [None]
        self.__counts = [None]
        self.__rank = [None]

        for i in range(1, sets+1):
            self.__parent.append(i)
            self.__movements.append(i)
            self.__rank.append(0)
            self.__counts.append(1)

    def find_set(self, item_number: int) -> int:
        '''
        Finds the representative node of the set containing 'item_number'

        Args:
            item_number: set number
        Returns:
            Number of representative set
        '''
        item_number = self.__movements[item_number]
        if self.__parent[item_number] == item_number:
            return item_number
        else:
            self.__parent[item_number] = self.find_set(self.__parent[item_number])
            return self.__parent[item_number]

    def is_same_set(self, first_item:int, second_item:int) -> bool:
        '''
        Checks if 2 items belong to the same set

        Args:
            first_item: First Set
            second_item: Second Set
        Returns:
            True if first_item and second_item belong to the same set
        '''
        return self.find_set(first_item) == self.find_set(second_item)

    def union(self, first_item:int, second_item:int) -> None:
        '''
        Unions 2 sets together

        Args:
            first_item: First Set
            second_item: Second Set
        '''
        if not self.is_same_set(first_item, second_item):
            rep_x = self.find_set(first_item)
            rep_y = self.find_set(second_item)

            if self.__rank[rep_x] > self.__rank[rep_y]:
                self.__parent[rep_y] = rep_x
                self.__counts[rep_x] += self.__counts[rep_y]
            else:
                self.__parent[rep_x] = rep_y
                self.__counts[rep_y] += self.__counts[rep_x]
                if self.__rank[rep_x] == self.__rank[rep_y]:
                    self.__rank[rep_y] += 1

    def move(self, item_number:int, destination_number:int) -> None:
        '''
        Move an item from its original set to a new set

        Args:
            item_number: Original set number
            destination_number: New set number
        '''
        root_from = self.find_set(self.__movements[item_number])
        root_to = self.find_set(self.__movements[destination_number])
        if (root_from != root_to):
            self.__counts[root_from] -= 1
            self.__counts[root_to] += 1
            self.__movements[item_number] = root_to

    def count_items(self, item_number: int) -> int:
        '''
        Count the number of items in a particular set

        Args:
            item_number: Set number

        Returns:
            Total number of items in that set
        '''
        rep = self.find_set(self.__movements[item_number])
        return self.__counts[rep]


from __future__ import annotations
from typing import Any


class _BinaryNode:
    '''
    INTERNAL CLASS. Basic building block for constructing an AVL tree.
    '''
    def __init__(self, item) -> None:
        self.__left_node = None
        self.__right_node = None
        self.__parent = None
        self.__item = item
        self.__size = 1
        self.__height = 0

    def update_size(self) -> int:
        '''
        Updates the total number of binary nodes under this node.
        Returns:
            The updated size
        '''
        if self.__left_node == None and self.__right_node == None:
            self.__size = 1
        elif self.__left_node == None:
            self.__size = 1 + self.__right_node.get_size()
        elif self.__right_node == None:
            self.__size = 1 + self.__left_node.get_size()
        else:
            self.__size = 1 + self.__left_node.get_size() + self.__right_node.get_size()
        return self.__size

    def update_height(self) -> int:
        '''
        Updates the new height of the binary nodes under this node.
        Returns:
            The updated height
        '''
        if self.__left_node == None and self.__right_node == None:
            self.__height = 0
        elif self.__left_node == None:
            self.__height = 1 + self.__right_node.get_height()
        elif self.__right_node == None:
            self.__height = 1 + self.__left_node.get_height()
        else:
            self.__height = 1 + \
                max(self.__left_node.get_height(), self.__right_node.get_height())
        return self.__height

    def __eq__(self, other: _BinaryNode) -> bool:
        '''
        Checks if two binary nodes are equivalent.
        Returns:
            Boolean of whether the 2 objects are equal.
        '''
        if not isinstance(other, _BinaryNode):
            return False
        return self.__item == other.get_item()

    # Getters and Setters

    def get_item(self):
        return self.__item

    def get_left(self):
        return self.__left_node

    def get_right(self):
        return self.__right_node

    def get_parent(self):
        return self.__parent

    def get_size(self):
        return self.__size

    def get_height(self):
        return self.__height

    def set_left(self, new_node: _BinaryNode) -> None:
        self.__left_node = new_node

    def set_right(self, new_node: _BinaryNode) -> None:
        self.__right_node = new_node

    def set_parent(self, new_node: _BinaryNode) -> None:
        self.__parent = new_node

    def set_item(self, new_item: Any) -> None:
        self.__item = new_item



from __future__ import annotations
from typing import Any, Callable
from .binary_node import _BinaryNode


class Avl:
    '''
    This data structure represents a self balancing binary tree
    '''

    def __init__(self, comparator: Callable = lambda x, y: x >= y) -> None:
        '''
        Args:
            comparator: Function which is used to decide the relative ordering of elements
        Comparators should be defined in the following format: 

        (x, y) => x>=y
        '''
        self.__root = None
        self.comparator = comparator

    def __helper_search(self, curr_node: _BinaryNode, item) -> _BinaryNode:
        '''
        INTERNAL FUNCTION: Helper function to find a particular item in the tree 

        Args:
            curr_node: The current Node to perform the search on
            item: The item that is to be found
        Returns:
            None or The Node which contains the item
        '''
        if curr_node is None:
            return None
        elif curr_node.get_item() == item:
            return curr_node
        elif self.comparator(curr_node.get_item(), item):
            return self.__helper_search(curr_node.get_left(), item)
        else:
            return self.__helper_search(curr_node.get_right(), item)

    def __balance_factor(self, curr_node: _BinaryNode) -> int:
        '''
        INTERNAL FUNCTION: Function to find the balance factor of a particular node 

        Args:
            curr_node: The node to get the balance factor
        Returns:
            The balance factor
        '''
        left = -1 if curr_node.get_left() is None else curr_node.get_left().get_height()
        right = -1 if curr_node.get_right() is None else curr_node.get_right().get_height()
        return left - right

    def __rotate_left(self, curr_node: _BinaryNode) -> _BinaryNode:
        '''
        INTERNAL FUNCTION: Perform a left rotation about the current node 

        Args:
            curr_node: The node to perform a left rotation
        Returns:
            The new pivot node
        '''
        rc = curr_node.get_right()
        if rc is None:
            return None
        rc.set_parent(curr_node.get_parent())
        if rc.get_parent() is None:
            self.__root = rc
        curr_node.set_parent(rc)
        curr_node.set_right(rc.get_left())
        if rc.get_left() is not None:
            rc.get_left().set_parent(curr_node)
        rc.set_left(curr_node)
        curr_node.update_size()
        curr_node.update_height()
        rc.update_size()
        rc.update_height()
        return rc

    def __rotate_right(self, curr_node: _BinaryNode) -> _BinaryNode:
        '''
        INTERNAL FUNCTION: Perform a right rotation about the current node 

        Args:
            curr_node: The node to perform a right rotation
        Returns:
            The new pivot node
        '''
        lc = curr_node.get_left()
        if lc is None:
            return None
        lc.set_parent(curr_node.get_parent())
        if lc.get_parent() is None:
            self.__root = lc
        curr_node.set_parent(lc)
        curr_node.set_left(lc.get_right())
        if lc.get_right() is not None:
            lc.get_right().set_parent(curr_node)
        lc.set_right(curr_node)
        curr_node.update_size()
        curr_node.update_height()
        lc.update_size()
        lc.update_height()
        return lc

    def __balance(self, curr_node: _BinaryNode) -> _BinaryNode:
        '''
        INTERNAL FUNCTION: Checks for inbalance at a node and performs the necessary rotations 

        Args:
            curr_node: The node to balance
        Returns:
            The new pivot
        '''
        factor = self.__balance_factor(curr_node)
        if factor == 2:
            if self.__balance_factor(curr_node.get_left()) == -1:
                curr_node.setLeft(self.__rotate_left(curr_node.get_left()))
                curr_node = self.__rotate_right(curr_node)
            else:
                curr_node = self.__rotate_right(curr_node)
        elif factor == -2:
            if self.__balance_factor(curr_node.get_right()) == 1:
                curr_node.set_right(self.__rotate_right(curr_node.get_right()))
                curr_node = self.__rotate_left(curr_node)
            else:
                curr_node = self.__rotate_left(curr_node)

        return curr_node

    def search(self, item: Any) -> Any:
        '''
        Search for an item in the tree 

        Args:
            item: The item to search for
        Returns:
            None or item
        '''
        res = self.__helper_search(self.__root, item)
        return None if res is None else res.get_item()

    def insert(self, item: Any) -> None:
        '''
        Insert an element into the tree 

        Args:
            item: The item to be inserted into the tree
        '''
        def helper(curr_node: _BinaryNode, item: Any) -> _BinaryNode:
            if curr_node is None:
                return _BinaryNode(item)
            elif self.comparator(curr_node.get_item(), item):
                curr_node.set_left(helper(curr_node.get_left(), item))
                curr_node.get_left().set_parent(curr_node)
            else:
                curr_node.set_right(helper(curr_node.get_right(), item))
                curr_node.get_right().set_parent(curr_node)

            if curr_node is not None:
                curr_node.update_height()
                curr_node.update_size()
                curr_node = self.__balance(curr_node)

            return curr_node

        self.__root = helper(self.__root, item)

    def delete(self, item: Any) -> None:
        '''
        Delete an element from the tree 

        Args:
            item: The item to be deleted from the tree
        '''
        def helper_delete(curr_node: _BinaryNode, item: Any) -> _BinaryNode:
            if curr_node is None:
                return curr_node
            elif curr_node.get_item() == item:
                if curr_node.get_left() is None and curr_node.get_right() is None:
                    curr_node = None
                elif curr_node.get_left() is None:
                    curr_node.get_right().set_parent(curr_node.get_parent())
                    curr_node = curr_node.get_right()
                elif curr_node.get_right() is None:
                    curr_node.get_left().set_parent(curr_node.get_parent())
                    curr_node = curr_node.get_left()
                else:
                    item_successor = self.successor(item)
                    curr_node.set_item(item_successor)
                    curr_node.set_right(helper_delete(
                        curr_node.get_right(), item_successor))
            elif self.comparator(curr_node.get_item(), item):
                curr_node.set_left(helper_delete(curr_node.get_left(), item))
            else:
                curr_node.set_right(helper_delete(curr_node.get_right(), item))

            if curr_node is not None:
                curr_node.update_height()
                curr_node.update_size()
                curr_node = self.__balance(curr_node)

            return curr_node

        self.__root = helper_delete(self.__root, item)

    def min(self) -> Any:
        '''
        Find the smallest element in the tree 

        Returns:
            The item which is the minimum
        '''
        curr_node = self.__root
        while curr_node.get_left() is not None:
            curr_node = curr_node.get_left()
        return curr_node.get_item()

    def max(self) -> Any:
        '''
        Find the largest element in the tree 

        Returns:
            The item which is the maximum
        '''
        curr_node = self.__root
        while curr_node.get_right() is not None:
            curr_node = curr_node.get_right()
        return curr_node.get_item()

    def successor(self, item: Any) -> Any:
        '''
        Finds the next biggest element after "item" 

        Args:
            item: To find the successor of this item
        Returns:
            The successor of the item
        '''
        def successor_helper(curr_node: _BinaryNode) -> _BinaryNode:
            if curr_node.get_right() is not None:
                temp_node = curr_node.get_right()
                while temp_node.get_left() is not None:
                    curr_node = curr_node.get_left()
                return temp_node
            else:
                curr = curr_node
                parent = curr_node.get_parent()
                while parent is not None and curr == parent.get_right():
                    curr = parent
                    parent = parent.get_parent()
                return parent

        target = self.__helper_search(self.__root, item)
        res = successor_helper(target)
        return None if res is None else res.get_item()

    def predecessor(self, item: Any) -> Any:
        '''
        Finds the next smallest element after "item" 

        Args:
            item: To find the predecessor of this item
        Returns:
            The predecessor of the item
        '''
        def predecessor_helper(curr_node: _BinaryNode) -> _BinaryNode:
            if curr_node.get_left() is not None:
                temp_node = curr_node.get_left()
                while temp_node.get_right() is not None:
                    curr_node = curr_node.get_right()
                return temp_node
            else:
                curr = curr_node
                parent = curr_node.get_parent()
                while parent is not None and curr == parent.get_left():
                    curr = parent
                    parent = parent.get_parent()
                return parent

        target = self.__helper_search(self.__root, item)
        res = predecessor_helper(target)
        return None if res is None else res.get_item()

    def rank(self, item: Any) -> int:
        '''
        Finds the rank of a particular item 

        Args:
            item: To find the rank of this item
        Returns:
            The item's rank
        '''
        def rank_helper(curr_node: _BinaryNode, item: Any) -> int:
            if curr_node == None:
                return -1
            if curr_node.get_item() == item:
                return curr_node.get_left().get_size() + 1 if curr_node.get_left() is not None else 1
            elif self.comparator(curr_node.get_item(), item):
                return rank_helper(curr_node.get_left(), item)
            else:
                left = curr_node.get_left().get_size() + 1 if curr_node.get_left() is not None else 1
                right = rank_helper(curr_node.get_right(), item)
                res = -1 if right == -1 else left + right
                return res

        return rank_helper(self.__root, item)

    def select(self, rank: int) -> Any:
        '''
        Given the rank, find the associated item 

        Args:
            rank: The rank of the item
        Returns:
            The item
        '''
        def select_helper(curr_node: _BinaryNode, rank: int) -> _BinaryNode:
            if curr_node is None:
                return None
            left_size = curr_node.get_left().get_size() + 1 if curr_node.get_left() is not None else 1
            if left_size == rank:
                return curr_node
            elif left_size > rank:
                return select_helper(curr_node.get_left(), rank)
            else:
                return select_helper(curr_node.get_right(), rank - left_size)

        res = select_helper(self.__root, rank)

        return res.get_item() if res is not None else None

    def in_order(self) -> list:
        '''
        Conduct an in order traversal of the tree and output the result in a list 

        Returns:
            list of in order sequence
        '''
        def in_order_helper(curr_node: _BinaryNode) -> list:
            if curr_node == None:
                return []
            return in_order_helper(curr_node.get_left()) + [curr_node.get_item()] + in_order_helper(curr_node.get_right())

        return in_order_helper(self.__root)

    def pre_order(self) -> list:
        '''
        Conduct a pre order traversal of the tree and output the result in a list 

        Returns:
            list of pre order sequence
        '''
        def pre_order_helper(curr_node: _BinaryNode) -> list:
            if curr_node == None:
                return []
            return [curr_node.get_item()] + pre_order_helper(curr_node.get_left()) + pre_order_helper(curr_node.get_right())

        return pre_order_helper(self.__root)

    def post_order(self) -> list:
        '''
        Conduct a post order traversal of the tree and output the result in a list 

        Returns:
            list of post order sequence
        '''
        def post_order_helper(curr_node: _BinaryNode) -> list:
            if curr_node == None:
                return []
            return post_order_helper(curr_node.get_left()) + post_order_helper(curr_node.get_right()) + [curr_node.get_item()]

        return post_order_helper(self.__root)

    def print(self, type: int = 1) -> None:
        '''
        3 Types of prints available: 

        In order Traversal (type = 1) 

        Pre order Traversal (type = 2) 

        Post order Traversal (type = 3) 


        Args:
            type: see above
        '''
        if type == 1:
            print(self.in_order())
        elif type == 2:
            print(self.pre_order())
        elif type == 3:
            print(self.post_order())



from __future__ import annotations
from typing import Any

class _Node:
    '''
    Basic Building Block of a linkedList
    Class Parameters:
        item: item to store in the node
        nextNode: Reference pointer to the next node
    '''
    def __init__(self, item:Any, next_node:_Node=None):
        self.__item = item
        self.__next_node = next_node

    def get_item(self) -> Any:
        return self.__item

    def next(self) -> _Node:
        return self.__next_node

    def set_item(self, new_item: Any) -> _Node:
        self.__item = new_item
        return self

    def set_next(self, new_node:_Node) -> _Node:
        self.__next_node = new_node
        return self


'''
List of sorting algorithms implemented:
    1. Bubble Sort (Normal / Optimised)
    2. Selection Sort
    3. Insertion Sort
    4. Merge Sort
    5. Quick Sort
    6. Radix Sort
'''
from typing import Callable, Generator
from .private_funcs import swap, merge, partition

def bubble_sort(lst:list, key:Callable = lambda x:x, visualise:bool = False, animate:bool = False) -> Generator:
    '''
    Bubble sort is in place & stable

    Args: 
        lst: Takes in a list 
        key: custom sorting key
        visualise: print lst after every loop
        animate: used for visualising the animation
    '''
    yield lst.copy()
    for i in range(len(lst) - 1):
        for j in range(len(lst) - 1):
            if key(lst[j]) > key(lst[j+1]):
                swap(lst, j, j+1)
                if animate:
                    yield lst.copy()

        if visualise:
            yield lst.copy()
        
def bubble_sort_optimised(lst:list, key:Callable = lambda x:x, visualise:bool = False, animate:bool = False) -> Generator:
    '''
    Optimised Bubble sort terminates when there are no swaps happening between elements

    Args: 
        lst: Takes in a list 
        key: custom sorting key
        visualise: print lst after every loop
        animate: used for visualising the animation
    '''
    yield lst.copy()
    for i in range(len(lst) - 1):
        no_swaps = True
        for j in range(len(lst) - 1):
            if key(lst[j]) > key(lst[j+1]):
                swap(lst, j, j+1)
                no_swaps = False
                if animate:
                    yield lst.copy()

        if visualise:
            yield lst.copy()

        if no_swaps:
            return

def selection_sort(lst:list, key:Callable = lambda x:x, visualise:bool = False, animate:bool = False) -> Generator:
    '''
    Selection sort is in place & not stable

    Args: 
        lst: Takes in a list 
        key: custom sorting key
        visualise: print lst after every loop
        animate: used for visualising the animation
    '''
    yield lst.copy()
    for i in range(len(lst)-1, 0, -1):
        largest = i
        for j in range(i):
            if key(lst[j]) > key(lst[largest]):
                largest = j
        swap(lst, i, largest)
        if animate or visualise:
            yield lst.copy()

def insertion_sort(lst:list, key:Callable = lambda x:x, visualise:bool = False, animate:bool = False) -> Generator:
    '''
    Insertion sort is in place & stable

    Args: 
        lst: Takes in a list 
        key: custom sorting key
        visualise: print lst after every loop
        animate: used for visualising the animation
    '''
    yield lst.copy()
    for idx in range(1, len(lst)):
        next = lst[idx]
        j = idx - 1
        while j >= 0 and key(lst[j]) > key(next):
            lst[j+1] = lst[j]
            j -= 1
        lst[j+1] = next
        if animate or visualise:
            yield lst.copy()

def merge_sort(lst:list, key:Callable = lambda x:x, visualise:bool = False, animate:bool = False) -> list:
    '''
    merge sort is not in place & stable

    Args: 
        lst: Takes in a list 
        key: custom sorting key
        visualise: print lst after every loop
        animate: used for visualising the animation
    '''
    gen = __merge_sort_helper(lst, 0, len(lst)-1, [lst.copy()], key)
    if visualise or animate:
        return gen
    return []

def __merge_sort_helper(lst:list, start:int, end:int, steps:list, key:Callable = lambda x:x) -> list:
    '''
    Merge sort recursive helper function. Divide and conquer method

    Args: 
        lst: Takes in a list 
        start: starting index to sort
        end: last index to sort
        steps: A list of state changes
        key: custom sorting key
    '''
    if (start < end):
        __merge_sort_helper(lst, start, (start + end)//2, steps, key)
        __merge_sort_helper(lst, 1+(start + end)//2, end, steps, key)
        merge(lst, start, (start + end)//2, end, key)
        steps.append(lst.copy())
    return steps

def quick_sort(lst:list, key:Callable = lambda x:x, visualise:bool = False, animate:bool = False) -> list:
    '''
    Quick sort is in place & not stable

    Args: 
        lst: Takes in a list 
        key: custom sorting key
        visualise: print lst after every loop
        animate: used for visualising the animation
    '''
    gen = __quick_sort_helper(lst, 0, len(lst)-1, [lst.copy()], key)
    if visualise or animate:
        return gen
    return []

def __quick_sort_helper(lst:list, start:int, end:int, steps:list, key:Callable = lambda x:x):
    '''
    Quick sort recursive helper function
    
    Args: 
        lst: Takes in a list 
        start: starting index to sort
        end: last index to sort
        steps: A list of state changes
        key: custom sorting key
    '''
    if (start < end):
        pivot = partition(lst, start, end, key)
        steps.append(lst.copy())
        __quick_sort_helper(lst, start, pivot-1, steps, key)
        __quick_sort_helper(lst, pivot+1, end, steps, key)
    return steps


import matplotlib.pyplot as plt
import matplotlib.animation as anim

def Show(lst, callable, speed = 0.5, sortKey = lambda x:x):
    '''
    A function to visualise sorting animations
    Args: 
        lst: Takes in a list 
        callable: Sorting function 
        Speed: Animation in seconds (Default 0.5 seconds)
        sortKey: sort by what (Default original value)
    '''
    epochs = [-2]
    frames = list(callable(lst, animate=True, key=sortKey))
    fig, ax = plt.subplots()

    def add_labels(lst):
        for i in range(len(lst)):
            plt.text(i, lst[i], lst[i], ha = 'center')

    def update_plot(x, rec, epochs, frames):
        ax.texts = []
        epochs[0] += 1
        for rec, val in zip(rec, x):
            rec.set_height(val)
        add_labels(frames[epochs[0]])

    bar_rec = ax.bar(range(len(lst)), frames[0], align='center')
    add_labels(frames[0])
    _ = anim.FuncAnimation(fig, func=update_plot, fargs=(bar_rec, epochs, frames), frames=frames, interval=speed*1000, repeat=False)
    plt.show()

def Print(lst, callable, sortKey = lambda x:x):
    '''
    A function to print the current state of list after every iteration
    Args: 
        lst: Takes in a list 
        callable: Sorting function 
        sortKey: sort by what
    '''
    for i in callable(lst, visualise=True, key=sortKey):
        print(i)


from __future__ import annotations
from typing import Any, Callable
from .linked_list import LinkedList
from .node import _Node

class TailedLinkedList(LinkedList):
    '''
    Tailed Linked List
    '''
    def __init__(self, items=None):
        '''
        Args:
            item: either None, 1 item, or a list/tuple of items
        '''
        self.__tail = None
        super().__init__(items)

    def add_front(self, item:Any) -> TailedLinkedList:
        '''
        Adds the item to the front of the TailedLinkedList

        Args:
            item: Item to be added at the front of the TailedLinkedList
        Returns:
            The updated TailedLinkedList
        '''
        new_node = _Node(item, self._LinkedList__head)
        self._LinkedList__head = new_node
        if self.__tail == None:
            self.__tail = new_node
        self._LinkedList__size += 1
        return self

    def add_back(self, item:Any) -> TailedLinkedList:
        '''
        Adds the item to the back of the TailedLinkedList

        Args:
            item: Item to be added at the back of the TailedLinkedList
        Returns:
            The updated TailedLinkedList
        '''
        new_node = _Node(item)

        if self._LinkedList__head == None:
            self._LinkedList__head = new_node
            self.__tail = new_node
        else:
            curr_node = self._LinkedList__head
            while curr_node.next() != None:
                curr_node = curr_node.next()
            curr_node.set_next(new_node)
            self.__tail = new_node
        self._LinkedList__size += 1
        return self

    def remove_front(self) -> TailedLinkedList:
        '''
        Remove the item at the front of the TailedLinkedList

        Returns:
            The updated TailedLinkedList
        '''
        if self._LinkedList__size == 1:
            self.__tail = None
        return super().remove_front()

    def remove_back(self) -> TailedLinkedList:
        '''
        Remove the item at the back of the TailedLinkedList

        Returns:
            The updated TailedLinkedList
        '''
        if self._LinkedList__size == 0:
            return self
        elif self._LinkedList__size == 1:
            self._LinkedList__head = None
            self.__tail = None
            self._LinkedList__size -= 1
        else:
            curr_node = self._LinkedList__head
            for _ in range(self._LinkedList__size - 2):
                curr_node = curr_node.next()
            self.__tail = curr_node
            curr_node.set_next(None)
            self._LinkedList__size -= 1
        return self

    def remove_at_index(self, index) -> TailedLinkedList:
        '''
        Remove an item from a specified index 

        Args:
            index: Index of item to remove
        Returns:
            The updated TailedLinkedList
        '''
        if self._LinkedList__size == 1:
            self.__tail = None
        return super().remove_at_index(index)

    def map(self, func:Callable[[Any], Any]) -> TailedLinkedList:
        '''
        Maps the current LinkedList to a function. Returns a new TailedLinkedList

        Args:
            func: A function to map the values by
        Returns:
            A new TailedLinkedList containing the mapped values
        '''
        new_ll = TailedLinkedList()
        curr_node = self._LinkedList__head

        while curr_node != None:
            new_ll.add_back(func(curr_node.get_item()))
            curr_node = curr_node.next()

        return new_ll

    def filter(self, func:Callable[[Any], Any]) -> TailedLinkedList:
        '''
        Filter the Linkedlist based on a function. Returns a new TailedLinkedlist

        Args:
            func: A function to filter the values by
        Returns:
            A new TailedLinkedList containing the filtered values
        '''
        new_ll = TailedLinkedList()
        curr_node = self._LinkedList__head
        while curr_node != None:
            if func(curr_node.get_item()):
                new_ll.add_back(curr_node.get_item())
            curr_node = curr_node.next()

        return new_ll

    def get(self, index:int) -> int:
        '''
        Returns the item at a specified index

        Args:
            index: index of the item
        
        Returns:
            None if the LinkedList is empty else the item
        '''
        if index == self._LinkedList__size - 1:
            return None if self.__tail is None else self.__tail.get_item()
        return super().get(index)


from typing import Any
from jellybeans.structures import Queue

class Deque(Queue):
    '''
    Deque is an extension of Queue, allowing addition of elements at the front of the queue
    '''
    def __init__(self):
        super().__init__()

    def enqueue_front(self, item:Any) -> None:
        '''
        Add an item to the front of the queue
        
        Args:
            item: item to be added
        '''
        self._Queue__size += 1
        self._Queue__queue.add_front(item)


class _NotOneBasedIndexed(Exception):
    '''
    An exception that will be called when a Heap takes in a sequence which is not one base indexed.
    '''
    def __init__(self, message):
        super().__init__(message)


from typing import Any

class Stack:
    '''
    Stack is a LIFO Data structure
    '''
    def __init__(self) -> None:
        self.__size = 0
        self.__stack = []

    def is_empty(self) -> bool:
        '''
        Checks if there are elements in the stack

        Returns:
            True if stack has 0 elements
        '''
        return self.__size == 0

    def peek(self) -> Any:
        '''
        Returns:
            element at the top of the stack
        '''
        if self.is_empty():
            return None
        return self.__stack[self.__size - 1]

    def pop(self) -> Any:
        '''
        Returns and remove the element at the top of the stack

        Returns:
            element at the top of the stack
        '''
        if self.is_empty():
            raise IndexError("Popping an empty stack")
        self.__size -= 1
        return self.__stack.pop()

    def push(self, item:Any) -> None:
        '''
        Insert an item into the top of the stack

        Args:
            item: Item to be inserted
        '''
        self.__stack.append(item)
        self.__size += 1

    def __len__(self) -> int:
        '''
        Returns the size of the stack
        '''
        return self.__size

    def __str__(self) -> str:
        '''
        Visual Representation of the stack
        '''
        if self.__size == 0:
            return "[]"
        result = ""
        for idx in range(len(self.__stack) - 1, -1, -1):
            result += str(self.__stack[idx])
            if idx > 0:
                result += " <- "
        return result



class _GraphProperty(Exception):
    '''
    An exception that will be called when certain graph properties are being violated.
    '''
    def __init__(self, message):
        super().__init__(message)


from __future__ import annotations
from typing import Any, Callable
from .linked_list import LinkedList
from .doubly_node import _DoublyNode

class DoublyLinkedList(LinkedList):
    '''
    Doubly Linked List
    '''
    def __init__(self, items:Any=None):
        '''
        Args:
            item: either None, 1 item, or a list/tuple of items
        '''
        self.__tail = None
        super().__init__(items)

    def add_front(self, item:Any) -> LinkedList:
        '''
        Adds the item to the front of the DoublyLinkedList

        Args:
            item: Item to be added at the front of the DoublyLinkedList
        Returns:
            The updated DoublyLinkedList
        '''
        new_node = _DoublyNode(item, self._LinkedList__head)
        self._LinkedList__head = new_node
        if self._LinkedList__head.next() != None:
            new_node.next().set_prev(new_node)
        if self.__tail == None:
            self.__tail = new_node
        self._LinkedList__size += 1
        return self

    def add_back(self, item:Any) -> DoublyLinkedList:
        '''
        Adds the item to the back of the DoublyLinkedList

        Args:
            item: Item to be added at the back of the DoublyLinkedList
        Returns:
            The updated DoublyLinkedList
        '''
        new_node = _DoublyNode(item)

        if self._LinkedList__head == None:
            self._LinkedList__head = new_node
            self.__tail = new_node
        else:
            curr_node = self._LinkedList__head
            while curr_node.next() != None:
                curr_node = curr_node.next()
            new_node.set_prev(curr_node)
            curr_node.set_next(new_node)
            self.__tail = new_node
        self._LinkedList__size += 1
        return self
    
    def add_at_index(self, item:Any, index:int) -> DoublyLinkedList:
        '''
        Adds the item at a specified index of the DoublyLinkedList

        Args:
            item: Item to be added
            index: position to add the item
        Returns:
            The updated LinkedList
        '''
        if index < 0:
            raise IndexError("Linkedlist does not support negative indexing")
        elif index > self._LinkedList__size:
            raise IndexError("Invalid Index: Proposed index larger than size of linkedList")
        elif index == 0:
            self.add_front(item)
        elif index == self._LinkedList__size:
            self.add_back(item)
        else:
            curr_node = self._LinkedList__head
            for _ in range(index-1):
                curr_node = curr_node.next()
            new_node = _DoublyNode(item, curr_node.next(), curr_node)
            curr_node.next().set_prev(new_node)
            curr_node.set_next(new_node)
            self._LinkedList__size += 1
        return self

    def remove_front(self) -> DoublyLinkedList:
        '''
        Remove the item at the front of the linked list

        Returns:
            The updated DoublyLinkedList
        '''
        if self._LinkedList__size == 1:
            self.__tail = None
        
        if self._LinkedList__size > 1:
            self._LinkedList__head.next().set_prev(None)
        return super().remove_front()

    def remove_back(self) -> DoublyLinkedList:
        '''
        Remove the item at the back of the DoublyLinkedList
        
        Returns:
            The updated DoublyLinkedList
        '''
        if self._LinkedList__size == 0:
            return self
        elif self._LinkedList__size == 1:
            self._LinkedList__head = None
            self.__tail = None
            self._LinkedList__size -= 1
            return self
        self.__tail = self.__tail.prev()
        self.__tail.set_next(None)
        self._LinkedList__size -= 1
        return self

    def remove_at_index(self, index:int) -> DoublyLinkedList:
        '''
        Remove an item from a specified index 

        Args:
            index: Index of item to remove
        Returns:
            The updated DoublyLinkedList
        '''
        if index == 0:
            return self.remove_front()
        elif index == self._LinkedList__size - 1:
            return self.remove_back()
        elif index >= self._LinkedList__size:
            raise IndexError("Invalid Index: Proposed index larger than size of linkedList")
        elif index < 0:
            raise IndexError("Linkedlist does not support negative indexing")
        else:
            curr_node = self._LinkedList__head
            for i in range(index - 1):
                curr_node = curr_node.next()
            curr_node.set_next(curr_node.next().next())
            curr_node.next().set_prev(curr_node)
            self._LinkedList__size -= 1
            return self

    def invert(self) -> DoublyLinkedList:
        '''
        Reverses the LL. 
        
        Returns:
            New inversed DoublyLinkedList 
        '''
        new_ll = DoublyLinkedList()
        curr_node = self.__tail
        while curr_node != None:
            new_ll.add_back(curr_node.get_item())
            curr_node = curr_node.prev()
        return new_ll

    def map(self, func:Callable[[Any], Any]) -> DoublyLinkedList:
        '''
        Maps the current LinkedList to a function. Returns a new DoublyLinkedList

        Args:
            func: A function to map the values by
        Returns:
            A new DoublyLinkedList containing the mapped values
        '''
        new_ll = DoublyLinkedList()
        curr_node = self._LinkedList__head

        while curr_node != None:
            new_ll.add_back(func(curr_node.get_item()))
            curr_node = curr_node.next()

        return new_ll

    def filter(self, func:Callable[[Any], Any]) -> DoublyLinkedList:
        '''
        Filter the Linkedlist based on a function. Returns a new DoublyLinkedlist

        Args:
            func: A function to filter the values by
        Returns:
            A new DoublyLinkedList containing the filtered values
        '''
        new_ll = DoublyLinkedList()
        curr_node = self._LinkedList__head
        while curr_node != None:
            if func(curr_node.get_item()):
                new_ll.add_back(curr_node.get_item())
            curr_node = curr_node.next()

        return new_ll

    def get(self, index:int) -> int:
        '''
        Returns the item at a specified index

        Args:
            index: index of the item
        
        Returns:
            None if the LinkedList is empty else the item
        '''
        if index == self._LinkedList__size - 1:
            return None if self.__tail is None else self.__tail.get_item()
        return super().get(index)


from jellybeans.structures.linkedLists.linked_list import *
from jellybeans.structures.linkedLists.tail_linked_list import *
from jellybeans.structures.linkedLists.doubly_linked_list import *
from jellybeans.structures.stack.stack import *
from jellybeans.structures.queue.queue import *
from jellybeans.structures.queue.deque import *
from jellybeans.structures.heaps.priority_queue import *
from jellybeans.structures.ufds.ufds import *
from jellybeans.structures.avl.avl import *
from jellybeans.structures.graph.graph import *


from __future__ import annotations
from typing import Any, Callable, NewType, Union
from .node import _Node

class LinkedList:
    '''
    Single Linked List
    '''
    def __init__(self, item:Any = None):
        '''
        Args:
            item: either None, 1 item, or a list/tuple of items
        '''
        self.__head = None
        self.__size = 0
        if item == None:
            return
        
        if (type(item) == list or type(item) == tuple):
            for i in item:
                self.add_back(i)
        else:
            self.add_back(item)

    def add_front(self, item:Any) -> LinkedList:
        '''
        Adds the item to the front of the linkedList

        Args:
            item: Item to be added at the front of the linked list
        Returns:
            The updated LinkedList
        '''
        new_node = _Node(item, self.__head)
        self.__head = new_node
        self.__size += 1
        return self
        
    def add_back(self, item:Any) -> LinkedList:
        '''
        Adds the item to the back of the linkedlist

        Args:
            item: Item to be added at the back of the linked list
        Returns:
            The updated LinkedList
        '''
        new_node = _Node(item)

        if self.__head == None:
            self.__head = new_node
        else:
            curr_node = self.__head
            while curr_node.next() != None:
                curr_node = curr_node.next()
            curr_node.set_next(new_node)
        self.__size += 1
        return self
    
    def add_at_index(self, item:Any, index:int) -> LinkedList:
        '''
        Adds the item at a specified index of the linkedlist
        Args:
            item: Item to be added
            index: position to add the item
        Returns:
            The updated LinkedList
        '''
        if index < 0:
            raise IndexError("Linkedlist does not support negative indexing")
        elif index > self.__size:
            raise IndexError("Invalid Index: Proposed index larger than size of linkedList")
        elif index == 0:
            self.add_front(item)
        elif index == self.__size:
            self.add_back(item)
        else:
            curr_node = self.__head
            for _ in range(index-1):
                curr_node = curr_node.next()
            new_node = _Node(item, curr_node.next())
            curr_node.set_next(new_node)
            self.__size += 1
        return self

    def remove_front(self) -> LinkedList:
        '''
        Remove the item at the front of the linked list

        Returns:
            The updated LinkedList
        '''
        if self.__size == 0:
            return self
        self.__head = self.__head.next()
        self.__size -= 1
        return self

    def remove_back(self) -> LinkedList:
        '''
        Remove the item at the back of the linked list

        Returns:
            The updated LinkedList
        '''
        if self.__size == 0:
            return self
        elif self.__size == 1:
            self.__head = None
            self.__size -= 1
        else:
            curr_node = self.__head
            for _ in range(self.__size - 2):
                curr_node = curr_node.next()
            curr_node.set_next(None)
            self.__size -= 1
        return self

    def remove_at_index(self, index:int) -> LinkedList:
        '''
        Remove an item from a specified index 

        Args:
            index: Index of item to remove
        Returns:
            The updated LinkedList
        '''
        if index == 0:
            return self.remove_front()
        elif index == self.__size - 1:
            return self.remove_back()
        elif index >= self.__size:
            raise IndexError("Invalid Index: Proposed index larger than size of linkedList")
        elif index < 0:
            raise IndexError("Linkedlist does not support negative indexing")
        else:
            curr_node = self.__head
            for _ in range(index - 1):
                curr_node = curr_node.next()
            curr_node.set_next(curr_node.next().next())
            self.__size -= 1
            return self

    def update(self, index: int, new_item:Any) -> LinkedList:
        '''
        Update the value of an item at a specified index

        Args:
            index: index to be updated
            newItem: Item to be inserted into the linkedlist
        Returns:
            The updated LinkedList
        '''
        curr_node = self.__head
        if curr_node == None:
            return self
        elif index < 0:
            raise IndexError("Linkedlist does not support negative indexing")
        elif index >= self.__size:
            raise IndexError("Invalid Index: Proposed index larger than size of linkedList")
        
        while index > 0:
            curr_node = curr_node.next()
            index -= 1
        curr_node.set_item(new_item)
        return self

    def get(self, index: int) -> Union[None, int]:
        '''
        Returns the item at a specified index

        Args:
            index: index of the item
        
        Returns:
            None if the LinkedList is empty else the item
        '''
        curr_node = self.__head
        if curr_node == None:
            return None
        elif index < 0:
            raise IndexError("Linkedlist does not support negative indexing")
        elif index >= self.__size:
            raise IndexError("Invalid Index: Proposed index larger than size of linkedList")
        
        while index > 0:
            curr_node = curr_node.next()
            index -= 1
        return curr_node.get_item()

    def map(self, func: Callable[[Any], Any]) -> LinkedList:
        '''
        Maps the current LinkedList to a function. Returns a new LinkedList

        Args:
            func: A function to map the values by
        Returns:
            A new LinkedList containing the mapped values
        '''
        new_ll = LinkedList()
        curr_node = self.__head

        while curr_node != None:
            new_ll.add_back(func(curr_node.get_item()))
            curr_node = curr_node.next()

        return new_ll

    def filter(self, func: Callable[[Any], Any]) -> LinkedList:
        '''
        Filter the Linkedlist based on a function. Returns a new Linkedlist

        Args:
            func: A function to filter the values by
        Returns:
            A new LinkedList containing the filtered values
        '''
        new_ll = LinkedList()
        curr_node = self.__head
        while curr_node != None:
            if func(curr_node.get_item()):
                new_ll.add_back(curr_node.get_item())
            curr_node = curr_node.next()

        return new_ll

    def to_list(self) -> list:
        '''
        Converts the LinkedList to a list

        Returns:
            A list containing the values in the LinkedList
        '''
        curr_node = self.__head
        result = []
        while curr_node != None:
            result.append(curr_node.get_item())
            curr_node = curr_node.next()
        return result

    def __len__(self) -> int:
        '''
        Returns the size of the linkedlist
        '''
        return self.__size

    def __str__(self) -> str:
        '''
        String Representation of the linkedlist
        '''
        currNode = self.__head
        items = []

        if currNode == None:
            return "[]"
        
        while currNode != None:
            items.append(str(currNode.get_item()))
            currNode = currNode.next()
        return " => ".join(items)

    def __eq__(self, other:LinkedList) -> bool:
        '''
        Tells whether two LinkedLists are equal
        '''
        if not isinstance(other, LinkedList):
            return False
        elif self.__size != len(other):
            return False

        curr1 = self.__head
        curr2 = other.__head

        while curr1 != None and curr2 != None:
            if curr1.get_item() != curr2.get_item():
                return False
            curr1 = curr1.next()
            curr2 = curr2.next()
        return True


from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='jellybeans',
    version='1.0.5',
    license='MIT',
    author='Lee Jia Jun',
    author_email='jiajunlee2@gmail.com',
    description="A Data Structure Package",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=find_packages(exclude=['tests']),
    url='https://github.com/Jcheez/Jellybeans',
    keywords='Data-Structures',
    install_requires=[
        'matplotlib'
    ],
    python_requires=">=3.7"
)